/*
Copyright 2011 David Green <david4dev@gmail.com>
`startup.vala` is part of `plug-startup`
File Authors:
    David Green <david4dev@gmail.com>
This is free software licenced under the GPL version 3, see:
    file://./LICENCE
    http://www.gnu.org/licenses/gpl-3.0.html
    file:///usr/share/common-licences/GPL-3
*/
using GLib;
using Gtk;
using Granite.Widgets;
const string GETTEXT_PACKAGE = "plug-startup";

namespace Startup {

    public string __FILE__;
    public string __DIR__;
    public string __LOCALE_DIR__;

    /*Startup applications with the following commands will be filtered
    out by default. There should be a link button to show technical
    applications.*/
    public const string[] TECHNICAL = {
        "/usr/lib/at-spi/at-spi-registryd",
        "/usr/bin/gnome-keyring-daemon --start --components=pkcs11",
        "/usr/lib/gnome-settings-daemon/gnome-settings-daemon",
        "gsettings-data-convert",
        "/usr/lib/policykit-1-gnome/polkit-gnome-authentication-agent-1",
        "start-pulseaudio-x11",
        "/usr/bin/gnome-keyring-daemon --start --components=secrets",
        "/usr/bin/gnome-keyring-daemon --start --components=ssh",
        "zeitgeist-datahub",
        "xdg-user-dirs-gtk-update",
        "sh -c 'test -e /var/cache/jockey/check || exec jockey-gtk --check'",
        "/usr/lib/gnome-disk-utility/gdu-notification-daemon",
        "nm-applet --sm-disable",
        "sh -c '/usr/bin/nvidia-settings --load-config-only'",
        "/usr/lib/gnome-user-share/gnome-user-share",
        "gnome-power-manager"
    };
    /*Startup applications with the following commands will be filtered
    out always*/
    public const string[] BLACKLIST = {
        "echo PLUG_STARTUP_DELETED"
    };

    public string[] DEFAULT_LANGUAGE;
    public string DEFAULT_LANGUAGE_BEST;

    /*Helper functions for handling lists.*/

    public delegate bool Filter<G> (G item);

    public List<G> filter<G> (List<G> list, Filter f) {
        var filtered = new List<G> ();
        foreach (G item in list) {
            if (f(item)) {
                filtered.append(item);
            }
        }
        return filtered;
    }

    public delegate string StringMap<G> (G item);

    public List<G> map_string<G> (List<G> list, StringMap m) {
        var mapped = new List<G> ();
        foreach (G item in list) {
            mapped.append(m(item));
        }
        return mapped;
    }

    public List<string> string_array_to_list (string[] array) {
        var list = new List<string> ();
        foreach (string str in array) {
            list.append(str);
        }
        return list;
    }

    public static int alphabetical (Application a, Application b) {
        string sa = (a.name + a.comment).up();
        string sb = (b.name + b.comment).up();
        return GLib.strcmp(sa, sb);
    }

    /*Helper functions for handling the file system.*/

    public bool file_exists (string path) {
        return FileUtils.test(
            path,
            FileTest.EXISTS
        );
    }
    public Filter<string> FILE_EXISTS;

    public bool dir_exists (string dir) {
        return file_exists(dir) && FileUtils.test(
            dir,
            FileTest.IS_DIR
        );
    }
    public Filter<string> DIR_EXISTS;

    public string path_join (string path, string name) {
        if (path[path.length - 1] == Path.DIR_SEPARATOR) {
            return path + name;
        }
        return path + Path.DIR_SEPARATOR_S + name;
    }

    public bool is_desktop_file (string name) {
        var splt = name.split(".");
        return splt[splt.length - 1] == "desktop" && name != "desktop";
    }
    public Filter<string> IS_DESKTOP_FILE;

    public List<string> files_in_dir (string path) throws FileError {
        var files = new List<string> ();
        var dir = Dir.open(path);
        string? name = dir.read_name();
        while (name != null) {
            files.append(path_join(path, name));
            name = dir.read_name();
        }
        return files;
    }

    public string basename (string path) {
        var splt = path.split(Path.DIR_SEPARATOR_S);
        return splt[splt.length - 1];
    }
    public StringMap<string> BASENAME;

    public string dirname (string path) {
        if (dir_exists(path)) {
            return path;
        }
        var splt = path.split(Path.DIR_SEPARATOR_S);
        var n = splt[0:splt.length - 1];
        return string.joinv(Path.DIR_SEPARATOR_S, n);
    }
    public StringMap<string> DIRNAME;

    public void copy_to_user_autostart (string file) throws FileError {
        string contents;
        FileUtils.get_contents(file, out contents);
        var name = basename(file);
        if (is_desktop_file(name)) {
            var outfile = path_join(user_autostart_dir(), name);
            if (!file_exists(outfile)) {
                FileUtils.set_contents(outfile, contents);
            }
        }
    }

    /*
    Functions for locating startup application desktop entries;
    following the freedesktop.org spec:
      http://standards.freedesktop.org/autostart-spec/autostart-spec-0.5.html
    */

    public string get_autostart_dir (string config_dir) {
        return path_join(config_dir, "autostart");
    }
    public StringMap<string> GET_AUTOSTART_DIR;

    public List<string> system_autostart_dirs () {
        return filter<string>(
            map_string<string>(
                string_array_to_list(
                    Environment.get_system_config_dirs()
                ),
                GET_AUTOSTART_DIR
            ),
            DIR_EXISTS
        );
    }

    public string user_autostart_dir () {
        return get_autostart_dir(
            Environment.get_user_config_dir()
        );
    }

    public List<string> autostart_files () {
        List<string> basenames = new List<string> ();
        List<string> dirs = system_autostart_dirs();
        List<string> files = new List<string> ();
        dirs.prepend(user_autostart_dir());
        foreach (string dir in dirs) {
            try {
                foreach (string file in filter<string>(
                    files_in_dir(dir),
                    IS_DESKTOP_FILE
                )) {
                    var name = basename(file);
                    if (basenames.index(name) < 0) {
                        basenames.append(name);
                        files.append(file);
                    }
                }
            } catch (FileError e) {}
        }
        return files;
    }

    public List<string> user_autostart_files () throws FileError {
        return filter<string>(
            files_in_dir(user_autostart_dir()),
            IS_DESKTOP_FILE
        );
    }

    /*Creates a user copy of the system wide autostart applications to
    allow editing.*/
    public void make_user_copy () {
        foreach (string path in autostart_files()) {
            try {
                copy_to_user_autostart(path);
            } catch (FileError e) {
                print(
                    "ERROR - couldn't create user copy for `" + path +
                    "`: " + e.message + "\n"
                );
            }
        }
    }

    /*Represents a startup application.*/
    public class Application : Object {

        public string name;
        public string command;
        public string comment;
        public string icon;
        public bool hidden;
        public string path;

        public Application(
            string name="",
            string command="",
            string comment="",
            string icon="application-x-executable",
            bool hidden=false,
            string path=""
        ) {
            this.name = name;
            this.command = command;
            this.comment = comment;
            this.icon = icon;
            this.hidden = hidden;
            this.path = path;
        }

        /*generate a path to save this application if it is a newly
        created one.*/
        public string generate_path () {
            var cmd = command.split(" ");
            var x = "";
            var i = 0;
            var p = "";
            do {
                p = path_join(
                    user_autostart_dir(),
                    cmd[0] + x + ".desktop"
                );
                i += 1;
                x = "-" + i.to_string();
            } while (file_exists(p));
            return p;
        }

        public string get_locale_string (KeyFile kf, string key) {
            foreach (string l in DEFAULT_LANGUAGE) {
                try {
                    return kf.get_locale_string(
                        KeyFileDesktop.GROUP,
                        key,
                        l
                    );
                } catch (KeyFileError e) {}
            }
            try {
                return kf.get_string(
                    KeyFileDesktop.GROUP,
                    key
                );
            } catch (KeyFileError e) {
                return "";
            }
        }

        public void save_name (KeyFile kf) {
            kf.set_locale_string(
                KeyFileDesktop.GROUP,
                KeyFileDesktop.KEY_NAME,
                DEFAULT_LANGUAGE_BEST,
                name
            );
            kf.set_string(
                KeyFileDesktop.GROUP,
                KeyFileDesktop.KEY_NAME,
                name
            );
        }
        public void load_name (KeyFile kf) {
            name = get_locale_string(
                kf, KeyFileDesktop.KEY_NAME
            );
        }

        public void save_comment (KeyFile kf) {
            kf.set_locale_string(
                KeyFileDesktop.GROUP,
                KeyFileDesktop.KEY_COMMENT,
                DEFAULT_LANGUAGE_BEST,
                comment
            );
            kf.set_string(
                KeyFileDesktop.GROUP,
                KeyFileDesktop.KEY_COMMENT,
                comment
            );
        }
        public void load_comment (KeyFile kf) {
            comment = get_locale_string(
                kf, KeyFileDesktop.KEY_COMMENT
            );
        }

        public void save_command (KeyFile kf) {
            kf.set_string(
                KeyFileDesktop.GROUP,
                KeyFileDesktop.KEY_EXEC,
                command
            );
        }
        public void load_command (KeyFile kf) {
            try {
                command = kf.get_string(
                    KeyFileDesktop.GROUP,
                    KeyFileDesktop.KEY_EXEC
                );
            } catch (KeyFileError e) {
                command = "";
            }
        }

        public void save_icon (KeyFile kf) {
            kf.set_string(
                KeyFileDesktop.GROUP,
                KeyFileDesktop.KEY_ICON,
                icon
            );
        }
        public void load_icon (KeyFile kf) {
            try {
                icon = kf.get_string(
                    KeyFileDesktop.GROUP,
                    KeyFileDesktop.KEY_ICON
                );
            } catch (KeyFileError e) {
                icon = "";
            }
        }

        public void save_hidden (KeyFile kf) {
            kf.set_boolean(
                KeyFileDesktop.GROUP,
                KeyFileDesktop.KEY_HIDDEN,
                hidden
            );
            kf.set_boolean(
                KeyFileDesktop.GROUP,
                "X-GNOME-Autostart-enabled",
                !hidden
            );
        }
        public void load_hidden (KeyFile kf) {
            try {
                hidden = kf.get_boolean(
                    KeyFileDesktop.GROUP,
                    KeyFileDesktop.KEY_HIDDEN
                );
            } catch (KeyFileError e) {
                hidden = false;
            }
        }

        public void save_all (KeyFile kf) {
            save_name(kf);
            save_command(kf);
            save_comment(kf);
            save_icon(kf);
            save_hidden(kf);
        }
        public void load_all (KeyFile kf) {
            load_name(kf);
            load_command(kf);
            load_comment(kf);
            load_icon(kf);
            load_hidden(kf);
        }

        public void save () throws FileError {
            if (path == "" || path == null) {
                path = generate_path();
            }
            var kf = new KeyFile();
            try {
                kf.load_from_file(path, KeyFileFlags.KEEP_TRANSLATIONS);
            } catch (Error e) {}
            this.save_all(kf);
            FileUtils.set_contents(path, kf.to_data());
        }
        public void load () throws KeyFileError, FileError {
            var kf = new KeyFile();
            kf.load_from_file(path, KeyFileFlags.KEEP_TRANSLATIONS);
            this.load_all(kf);
        }
        public void remove () throws FileError {
            if (file_exists(path)) {
                hidden = true;
                command = "echo PLUG_STARTUP_DELETED";
                save();
            }
        }

        public static Application from_file (string f) throws KeyFileError, FileError {
            var app = new Application();
            app.path = f;
            app.load();
            return app;
        }

        public static List<Application> list_all () {
            var l = new GLib.List<Application> ();
            try {
                foreach (string f in user_autostart_files()) {
                    try {
                        l.append(Application.from_file(f));
                    } catch (Error e) {
                        print("ERROR: " + e.message + "\n");
                    }
                }
            } catch (FileError e) {
                print("ERROR: " + e.message + "\n");
            }
            return l;
        }

        public static bool not_blacklisted (Application app) {
            foreach (string cmd in BLACKLIST) {
                if (app.command == cmd) {
                    return false;
                }
            }
            return true;
        }
        public static Filter<Application> NOT_BLACKLISTED = not_blacklisted;

        public static bool not_technical (Application app) {
            foreach (string cmd in TECHNICAL) {
                if (app.command == cmd) {
                    return false;
                }
            }
            return true;
        }
        public static Filter<Application> NOT_TECHNICAL = not_technical;

        public static List<Application> list_not_blacklisted () {
            return filter<Application>(
                list_all(),
                NOT_BLACKLISTED
            );
        }

        public static List<Application> list_not_technical () {
            return filter<Application>(
                list_not_blacklisted(),
                NOT_TECHNICAL
            );
        }


    }

    public void init (string self_path) {
        __FILE__ = self_path;
        __DIR__ = dirname(__FILE__);
        __LOCALE_DIR__ = path_join(
            path_join(
                path_join(__DIR__, ".."),
                "share"
            ),
            "locale"
        );
        if (!dir_exists(__LOCALE_DIR__)) {
            __LOCALE_DIR__ = "/usr/share/locale";
        }
        Intl.setlocale(LocaleCategory.MESSAGES, null);
        Intl.bindtextdomain("plug-startup", __LOCALE_DIR__);
        DEFAULT_LANGUAGE = Intl.get_language_names();
        DEFAULT_LANGUAGE_BEST = DEFAULT_LANGUAGE[0];
        FILE_EXISTS = file_exists;
        DIR_EXISTS = dir_exists;
        IS_DESKTOP_FILE = is_desktop_file;
        BASENAME = basename;
        DIRNAME = dirname;
        GET_AUTOSTART_DIR = get_autostart_dir;
        make_user_copy();
    }

    public static const int VPADDING = 4;
    public static const int HPADDING = 12;

    public class ApplicationList : ScrolledWindow {
        ListStore list_store;
        TreeView tree_view;
        TreeSelection selection;
        public List<Application> app_list;

        enum Columns {
            TOGGLE,
            ICON,
            TEXT,
            N_COLUMNS
        }

        public ApplicationList () {
            app_list = Application.list_not_technical();

            list_store = new ListStore (
                Columns.N_COLUMNS,
                typeof(bool),
                typeof(string),
                typeof(string)
            );
            tree_view = new TreeView.with_model (list_store);
            tree_view.rules_hint = true;

            var toggle = new CellRendererToggle ();
            toggle.toggled.connect ((toggle, path) => {
                var tree_path = new TreePath.from_string (path);
                var index = int.parse(path);
                if (app_list.length() > index) {
                    var app = app_list.nth_data(index);
                    app.hidden = toggle.active;
                    try {
                        app.save();
                    } catch (Error e) {
                        print(
                            "Error saving application `" + app.name +
                            "` - " + e.message + "\n"
                        );
                    }
                }
                TreeIter iter;
                list_store.get_iter (out iter, tree_path);
                list_store.set (iter, Columns.TOGGLE, !toggle.active);
            });

            var column = new TreeViewColumn ();
            column.pack_start (toggle, false);
            column.add_attribute (toggle, "active", Columns.TOGGLE);
            tree_view.append_column (column);

            var icon = new CellRendererPixbuf();
            icon.stock_size = IconSize.DIALOG;
            column = new TreeViewColumn ();
            column.pack_start (icon, true);
            column.add_attribute (icon, "icon_name", Columns.ICON);
            tree_view.append_column (column);

            var text = new CellRendererText ();
            column = new TreeViewColumn ();
            column.pack_start (text, true);
            column.add_attribute (text, "markup", Columns.TEXT);
            tree_view.append_column (column);

            tree_view.set_headers_visible (false);

            update_list();

            add(tree_view);

            selection = tree_view.get_selection();
            selection.set_mode(SelectionMode.SINGLE);
            selection.changed.connect(
                () => {
                    var app = get_selected();
                    if (!(app == null)) {
                        changed(app);
                    }
                }
            );
        }

        public Application? get_selected () {
            TreeModel model;
            TreeIter iter;
            if (selection.get_selected(out model, out iter)) {
                return app_list.nth_data(
                    int.parse(
                        model.get_path(iter).to_string()
                    ) //there must be a better way to do this
                );
            } else {
                return null;
            }
        }

        public signal void changed (Application app);

        public void update_list () {
            app_list.sort(alphabetical);
            list_store.clear();
            TreeIter iter;
            foreach (Application app in app_list) {
                list_store.append(out iter);
                list_store.set(
                    iter,
                    Columns.TOGGLE, !(app.hidden),
                    Columns.ICON, app.icon,
                    Columns.TEXT, "<b>" + app.name + "</b>\n" +
                        app.comment
                );
            }
        }

        public void show_technical () {
            app_list = Application.list_not_blacklisted();
            update_list();
        }

        public void hide_technical () {
            app_list = Application.list_not_technical();
            update_list();
        }
    }

    public class TechnicalToggle : HBox {
        public LinkButton button_show;
        public LinkButton button_hide;
        public bool showing = false;
        public TechnicalToggle () {
            button_show = new LinkButton.with_label(
                "",
                _("show technical applications")
            );
            button_hide = new LinkButton.with_label(
                "",
                _("hide technical applications")
            );
            pack_end(button_show, false, false, HPADDING);
            pack_end(button_hide, false, false, HPADDING);
            button_hide.clicked.connect(toggle);
            button_show.clicked.connect(toggle);
            button_hide.hide();
            button_hide.no_show_all = true;
        }
        public signal void hid();
        public signal void shown();
        public void toggle () {
            showing = !showing;
            if (showing) {
                button_show.hide();
                button_hide.show();
                shown();
            } else {
                button_hide.hide();
                button_show.show();
                hid();
            }
        }
    }

    public class Editor : VBox {
        public Application app;
        public Button new_button;
        public HintedEntry name_entry;
        public HintedEntry command_entry;
        public HintedEntry comment_entry;
        public HintedEntry icon_entry;
        public Button save_button;
        public Button revert_button;
        public Button remove_button;
        public Image icon;
        public static VBox make_vbox (Widget[] widgets, bool expand=false, bool homo=false) {
            var vb = new VBox(homo, 0);
            foreach (Widget widget in widgets) {
                vb.pack_start(widget, expand, expand, VPADDING);
            }
            return vb;
        }
        public static HBox make_hbox (Widget[] widgets, bool expand=false, bool homo=false) {
            var hb = new HBox(homo, 0);
            foreach (Widget widget in widgets) {
                hb.pack_start(widget, expand, expand, HPADDING);
            }
            return hb;
        }
        public static HBox make_label (string str) {
            var l = new Label("<b>"+str+":</b>");
            l.use_markup = true;
            var hb = new HBox(false, 0);
            hb.pack_end(l, false, false, 0);
            return hb;
        }
        public static HintedEntry make_entry (string str) {
            return new HintedEntry(str);
        }
        public void create_buttons () {
            new_button = new Button.from_stock(Stock.NEW);
            save_button = new Button.from_stock(Stock.SAVE);
            revert_button = new Button.from_stock(Stock.REVERT_TO_SAVED);
            remove_button = new Button.from_stock(Stock.REMOVE);
        }
        public void create_entries () {
            name_entry = make_entry(_("Application name"));
            command_entry = make_entry(
                _("Command used to launch the application")
            );
            comment_entry = make_entry(
                _("Comment describing the application")
            );
            icon_entry = make_entry(_("Icon name"));
        }
        public void init_events () {
            name_entry.changed.connect(on_change);
            command_entry.changed.connect(on_change);
            comment_entry.changed.connect(on_change);
            icon_entry.changed.connect(on_change);
            new_button.clicked.connect(do_new);
            save_button.clicked.connect(do_save);
            revert_button.clicked.connect(do_revert);
            remove_button.clicked.connect(do_remove);
        }
        public Editor (Application app) {
            this.app = app;
            create_buttons();
            create_entries();
            init_events();
            icon = new Image.from_icon_name(
                "application-x-executable",
                IconSize.DIALOG
            );
            var new_button_and_icon = new VBox(false, 0);
            new_button_and_icon.pack_start(
                new_button,
                false,
                false,
                VPADDING
            );
            new_button_and_icon.pack_end(
                icon,
                false,
                false,
                VPADDING
            );
            var hb = make_hbox({
                new_button_and_icon,
                make_vbox({
                    make_label(_("Name")),
                    make_label(_("Command")),
                    make_label(_("Comment")),
                    make_label(_("Icon"))
                }, false, true)
            });
            hb.pack_start(
                make_vbox({
                    name_entry,
                    command_entry,
                    comment_entry,
                    icon_entry
                }, true, true),
                true, true, HPADDING
            );
            hb.pack_start(
                make_vbox(
                    {save_button, revert_button, remove_button},
                    false, true
                ),
                false, false, HPADDING
            );
            this.add(hb);
        }
        public void set_name (string name) {
            name_entry.set_text(name);
        }
        public void set_command (string command) {
            command_entry.set_text(command);
        }
        public void set_comment (string comment) {
            comment_entry.set_text(comment);
        }
        public void set_icon_image (string icon_name_or_path) {
            if (file_exists(icon_name_or_path)) {
                icon.set_from_file(icon_name_or_path);
            } else {
                icon.set_from_icon_name(
                    icon_name_or_path,
                    IconSize.DIALOG
                );
            }
        }
        public void set_icon (string icon_name_or_path) {
            icon_entry.set_text(icon_name_or_path);
            set_icon_image(icon_name_or_path);
        }
        public string get_name () {
            return name_entry.get_text();
        }
        public string get_command () {
            return command_entry.get_text();
        }
        public string get_comment () {
            return comment_entry.get_text();
        }
        public string get_icon () {
            return icon_entry.get_text();
        }
        public void on_change () {
            set_icon_image(get_icon());
            set_sensitivity();
        }
        public void load () {
            set_name(app.name);
            set_command(app.command);
            set_comment(app.comment);
            set_icon(app.icon);
        }
        public void save () {
            app.name = get_name();
            app.command = get_command();
            app.comment = get_comment();
            app.icon = get_icon();
            try {
                app.save();
            } catch (FileError e) {
                print(
                    "Couldn't save to `" + app.path + "` - " +
                    e.message + "\n"
                );
            }
        }
        public signal void updated ();
        public bool is_unchanged () {
            return app.name == get_name() &&
                app.command == get_command() &&
                app.comment == get_comment() &&
                app.icon == get_icon();
        }
        public bool is_changed () {
            return !is_unchanged();
        }
        public bool string_has_no_content (string str) {
            if (str.length == 0) {
                return true;
            }
            var i = 0;
            while (i < str.length) {
                switch (str[i]) {
                    case ' ':
                        break;
                    case '\t':
                        break;
                    default:
                        return false;
                }
                i += 1;
            }
            return true;
        }
        public bool is_valid () {
            return !(
                string_has_no_content(get_name()) ||
                string_has_no_content(get_command())
            );
        }
        public bool can_save () {
            return is_changed() && is_valid();
        }
        public void do_save () {
            save();
            updated();
        }
        public bool can_remove () {
            return file_exists(app.path);
        }
        public void do_remove () {
            try {
                app.remove();
            } catch (FileError e) {
                print("Error removing app: " + e.message + "\n");
            }
            do_new();
            updated();
        }
        public bool can_revert () {
            return is_changed();
        }
        public void do_revert () {
            load();
        }
        public bool is_new () {
            var d = new Application();
            return get_name() == d.name &&
                get_command() == d.command &&
                get_comment() == d.comment &&
                get_icon() == d.icon;
        }
        public bool can_new () {
            return !is_new();
        }
        public void do_new () {
            app = new Application();
            load();
        }
        public void set_sensitivity () {
            save_button.sensitive = can_save();
            remove_button.sensitive = can_remove();
            revert_button.sensitive = can_revert();
            new_button.sensitive = can_new();
        }
    }

    public class Controller : VBox {
        public ApplicationList apps;
        public TechnicalToggle toggle;
        public Editor editor;
        public Controller () {
            apps = new ApplicationList();
            toggle = new TechnicalToggle();
            toggle.shown.connect(apps.show_technical);
            toggle.hid.connect(apps.hide_technical);
            editor = new Editor(new Application());
            editor.updated.connect(() => {
                if (toggle.showing) {
                    apps.show_technical();
                } else {
                    apps.hide_technical();
                }
            });
            apps.changed.connect(
                (app) => {
                    editor.app = app;
                    editor.load();
                    editor.set_sensitivity();
                }
            );
            pack_start(apps, true, true, 0);
            pack_start(toggle, false, false, VPADDING);
            pack_start(editor, false, false, VPADDING);
        }
    }

    public class MainWindow : Window {
        public MainWindow () {
            title = _("Startup Applications");
            set_default_icon_name("gnome-session-properties");
            add(new Controller());
            destroy.connect(Gtk.main_quit);
            resize(800, 600);
            show_all();
        }
    }

    public class Plug : Pantheon.Switchboard.Plug {
        public Plug () {
            var controller = new Controller();
            controller.show_all();
            add(controller);
        }
    }

    public static void main (string[] args) {
        Startup.init(args[0]);
        print(__FILE__ + "\n");
        print(__DIR__ + "\n");
        print(__LOCALE_DIR__ + "\n");
        Gtk.init(ref args);
        if (args.length > 1 && args[1] == "plug") {
            var plug = new Plug ();
            plug.register(_("Startup Applications"));
            plug.show();
        } else {
            new MainWindow();
        }
        Gtk.main();
    }

}
